# -*- coding: utf-8 -*-
#
# This file is part of Flask-AppExts
# Copyright (C) 2015 CERN.
#
# Flask-AppExts is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.

"""Bundle class with support for bower dependencies."""

from __future__ import absolute_import, print_function

from flask_assets import Bundle as BundleBase


class BowerBundle(BundleBase):
    """Bundle extension with a name and bower dependencies.

    The bower dependencies are used to generate a bower.json file.
    """

    def __init__(self, *contents, **options):
        """Initialize the named bundle.

        :param name: name of the bundle
        :type name: str
        :param bower: bower dependencies
        :type bower: dict
        """
        self.bower = options.pop("bower", {})
        super(BowerBundle, self).__init__(*contents, **options)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os
import shutil
import tempfile

import pytest
from flask import Flask
from flask_cli import FlaskCLI, ScriptInfo

from invenio_assets import InvenioAssets
from invenio_assets.bower import BowerBundle


@pytest.fixture()
def app():
    """Flask application fixture."""
    app = Flask(__name__)
    FlaskCLI(app)
    return app


@pytest.fixture()
def script_info(request):
    """Get ScriptInfo object for testing CLI."""
    initial_dir = os.getcwd()
    instance_path = tempfile.mkdtemp()
    app = Flask(__name__, instance_path=instance_path)
    FlaskCLI(app)
    InvenioAssets(app)
    os.chdir(instance_path)

    def teardown():
        shutil.rmtree(instance_path)
        os.chdir(initial_dir)

    request.addfinalizer(teardown)
    return ScriptInfo(create_app=lambda info: app)


@pytest.fixture()
def script_info_assets(request):
    """Get ScriptInfo object for testing CLI."""
    initial_dir = os.getcwd()
    instance_path = tempfile.mkdtemp()
    app = Flask(__name__, instance_path=instance_path)
    FlaskCLI(app)
    InvenioAssets(app)
    os.chdir(instance_path)

    class Ext(object):
        def __init__(self, app):
            assets = app.extensions['invenio-assets']
            assets.env.register('testbundle', BowerBundle(
                'test.css',
                output='testbundle.css'))
    static_dir = os.path.join(os.path.dirname(__file__), 'static')
    if not os.path.exists(static_dir):
        os.makedirs(static_dir)
    test_css = open(os.path.join(static_dir, 'test.css'), 'w+')
    test_css.write("Test")

    Ext(app)

    def teardown():
        shutil.rmtree(instance_path)
        os.chdir(initial_dir)

    request.addfinalizer(teardown)
    return ScriptInfo(create_app=lambda info: app)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Media asset management for Invenio."""

from __future__ import absolute_import, print_function

import pkg_resources
from flask_assets import Environment
from flask_collect import Collect

from .cli import assets as assets_cmd
from .cli import bower, collect


class InvenioAssets(object):
    """Invenio asset extension."""

    def __init__(self, app=None, entrypoint='invenio_assets.bundles',
                 **kwargs):
        """Extension initialization."""
        self.env = Environment()
        self.collect = Collect()
        self.entrypoint = entrypoint

        if app:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):
        """Initialize application object."""
        self.init_config(app)
        self.env.init_app(app)
        self.collect.init_app(app)
        self.init_cli(app.cli)
        if self.entrypoint:
            self.load_entrypoint(self.entrypoint)
        app.extensions['invenio-assets'] = self

    def init_config(self, app):
        """Initialize configuration."""
        app.config.setdefault("REQUIREJS_BASEURL", app.static_folder)
        app.config.setdefault('COLLECT_STATIC_ROOT', app.static_folder)
        app.config.setdefault('COLLECT_STORAGE', 'flask_collect.storage.link')

    def init_cli(self, cli):
        """Initialize CLI."""
        cli.add_command(assets_cmd)
        cli.add_command(bower)
        cli.add_command(collect)

    def load_entrypoint(self, entrypoint):
        """Load entrypoint."""
        for ep in pkg_resources.iter_entry_points(entrypoint):
            self.env.register(ep.name, ep.load())



# -*- coding: utf-8 -*-
#
# This file is part of Flask-AppExts
# Copyright (C) 2015 CERN.
#
# Flask-AppExts is free software; you can redistribute it and/or
# modify it under the terms of the Revised BSD License; see LICENSE
# file for more details.

"""Filters for webassets."""

from __future__ import absolute_import, print_function

import os

from flask import current_app
from webassets.filter import ExternalTool
from webassets.filter.requirejs import RequireJSFilter as RequireJSFilterBase


class RequireJSFilter(RequireJSFilterBase):
    """Optimize AMD-style modularized JavaScript into a single asset.

    Adds support for exclusion of files already in defined in other bundles.
    """

    def __init__(self, *args, **kwargs):
        """Initialize filter."""
        self.excluded_bundles = kwargs.pop('exclude', [])
        super(RequireJSFilter, self).__init__(*args, **kwargs)

    def setup(self):
        """Setup filter (only called when filter is actually used)."""
        super(RequireJSFilter, self).setup()

        excluded_files = []
        for bundle in self.excluded_bundles:
            excluded_files.extend(
                map(lambda f: os.path.splitext(f)[0],
                    bundle.contents)
            )

        if excluded_files:
            self.argv.append(
                "exclude={0}".format(",".join(excluded_files))
            )


class CleanCSSFilter(ExternalTool):
    """Minify css using cleancss.

    Implements opener capable of rebasing relative CSS URLs against
    ``COLLECT_STATIC_ROOT``.
    """

    name = 'cleancssurl'
    method = 'open'
    options = {
        'binary': 'CLEANCSS_BIN',
    }

    def setup(self):
        """Initialize filter just before it will be used."""
        super(CleanCSSFilter, self).setup()
        self.root = current_app.config.get('COLLECT_STATIC_ROOT')

    def open(self, out, source_path, **kw):
        """Open source."""
        self.subprocess(
            [self.binary or 'cleancss', '--root', self.root, source_path],
            out
        )

    def output(self, _in, out, **kw):
        """Output filtering."""
        self.subprocess([self.binary or 'cleancss'], out, _in)

    def input(self, _in, out, **kw):
        """Input filtering."""
        self.subprocess([self.binary or 'cleancss'], out, _in)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Media assets management for Invenio.

First make sure you have Flask application with Click support (meaning
Flask 1.0+ or alternatively use the Flask-CLI extension):

    >>> from flask import Flask
    >>> from flask_cli import FlaskCLI
    >>> app = Flask('myapp')
    >>> cli = FlaskCLI(app)

Next, initialize your extension:

    >>> from invenio_assets import InvenioAssets
    >>> assets = InvenioAssets(app)


Registering bundles
~~~~~~~~~~~~~~~~~~~
After having initialized the extension you can register bundles:

    >>> from webassets import Bundle
    >>> assets.env.register('mybundle', Bundle())
    <Bundle ...>


Command-line interface
~~~~~~~~~~~~~~~~~~~~~~
Invenio Assets installs three commands on your application:

 * ``assets`` - Assets building commands.
 * ``bower`` - Generation of bower.json.
 * ``collect`` - Collect static files.


.. code-block:: console

   $ inveniomanage bower -o bower.json
   $ bower install
   $ inveniomanage assets build
   $ inveniomanage collect

"""

from __future__ import absolute_import, print_function

from .bower import BowerBundle
from .ext import InvenioAssets
from .filters import CleanCSSFilter, RequireJSFilter
from .proxy import current_assets
from .version import __version__

__all__ = (
    '__version__', 'BowerBundle', 'CleanCSSFilter', 'InvenioAssets',
    'RequireJSFilter', 'current_assets')



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Test BowerBundle."""

from __future__ import absolute_import, print_function

import json
import os

from click.testing import CliRunner

from invenio_assets import BowerBundle
from invenio_assets.cli import bower


def test_init():
    """Test version import."""
    bundle = BowerBundle(
        'testfile.css',
        bower={
            "bootstrap": "3.0.0"
        }
    )
    assert bundle.bower == {"bootstrap": "3.0.0"}


def test_cli(script_info):
    """Test bower CLI."""
    deps = {"boostrap": "3.0.0"}

    runner = CliRunner()
    app = script_info.load_app()
    assets = app.extensions['invenio-assets']
    assets.env.register('test1', BowerBundle(bower=deps))
    assert len(assets.env) == 1

    expected = {
        "name": app.name,
        "dependencies": {
            "boostrap": "3.0.0",
        },
        "version": "",
    }

    # Test default output
    with runner.isolated_filesystem():
        result = runner.invoke(bower, obj=script_info)
        assert result.exit_code == 0

        filepath = os.path.join(app.instance_path, 'bower.json')
        assert os.path.exists(filepath)

        with open(filepath) as f:
            bower_json = json.loads(f.read())
        assert bower_json == expected

    # Test writing bower.json file.
    with runner.isolated_filesystem():
        result = runner.invoke(bower, ['-o', 'bower.json'], obj=script_info)
        assert result.exit_code == 0

        filepath = os.path.join('bower.json')
        assert os.path.exists(filepath)

        with open('bower.json') as f:
            bower_json = json.loads(f.read())
        assert bower_json == expected

    # Test merging a base another file.
    newdep = {'font-awesome': '4.0'}
    expected['dependencies'].update(newdep)
    with runner.isolated_filesystem():
        with open('base.json', 'wt') as f:
            f.write(json.dumps({'dependencies': newdep}))

        result = runner.invoke(
            bower, ['-i', 'base.json', '-o', 'bower.json'], obj=script_info)
        assert result.exit_code == 0

        filepath = os.path.join('bower.json')
        assert os.path.exists(filepath)

        with open('bower.json') as f:
            bower_json = json.loads(f.read())
        assert bower_json == expected



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Test Invenio Assets module."""

from __future__ import absolute_import, print_function

from mock import patch
from pkg_resources import EntryPoint
from webassets import Bundle

from invenio_assets import InvenioAssets


class MockEntryPoint(EntryPoint):
    """Mocking of entrypoint."""

    def load(self):
        """Mock load entry point."""
        return Bundle(self.name)


def _mock_entry_points(name):
    data = {
        'invenio_assets.bundles': [MockEntryPoint('demo.bundle0',
                                                  'demo.bundle0'),
                                   MockEntryPoint('demo.bundle1',
                                                  'demo.bundle1')],
    }
    names = data.keys() if name is None else [name]
    for key in names:
        for entry_point in data[key]:
            yield entry_point


def test_version():
    """Test version import."""
    from invenio_assets import __version__
    assert __version__


def test_init(app):
    """Test module initialization."""
    InvenioAssets(app)
    assets = app.extensions['invenio-assets']
    assert assets.env
    assert assets.collect
    assert app.config['REQUIREJS_BASEURL'] == app.static_folder
    assert app.config['COLLECT_STATIC_ROOT'] == app.static_folder


def test_init_post(app):
    """Test module initialization using init_app."""
    assets = InvenioAssets()
    assert assets.env
    assert assets.collect
    assert 'REQUIREJS_BASEURL' not in app.config
    assert 'COLLECT_STATIC_ROOT' not in app.config
    assets.init_app(app)
    assert 'REQUIREJS_BASEURL' in app.config
    assert 'COLLECT_STATIC_ROOT' in app.config


def test_init_cli(app):
    """Test cli registration."""
    assert len(app.cli.commands) == 0
    InvenioAssets(app)
    assert len(app.cli.commands) == 3


def test_assets_usage(app):
    """Test assets usage."""
    class Ext(object):
        def __init__(self, app):
            assets = app.extensions['invenio-assets']
            assets.env.register('testbundle', Bundle())

    assets = InvenioAssets(app)
    Ext(app)
    assert len(assets.env) == 1


@patch('pkg_resources.iter_entry_points', _mock_entry_points)
def test_assets_usage_entrypoints(app):
    """Test of entrypoints using mocking."""
    assets = InvenioAssets(app)
    assert len(assets.env) == 2



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Test Invenio Assets module."""

from __future__ import absolute_import, print_function

from flask import Flask
from flask_cli import FlaskCLI
from webassets.test import TempEnvironmentHelper

from invenio_assets import InvenioAssets
from invenio_assets.filters import CleanCSSFilter, RequireJSFilter


class TestInvenioAssetsRequireJSFilter(TempEnvironmentHelper):
    """"Test class checking the behaviour of JSRequire filter."""
    default_files = {
        'foo.js': """
        more();
        """,
        'bar.js': """
        function foo(bar) {
            document.write ( bar );
        }
        """,
    }

    def test_require_js(self):
        """Test method of RequireJS filter."""
        bundle_0 = self.mkbundle('bar.js',
                                 filters=RequireJSFilter(),
                                 output='out_0.js')
        bundle_0.build()

        bundle_1 = self.mkbundle('foo.js', 'bar.js',
                                 filters=RequireJSFilter(exclude=[bundle_0]),
                                 output='out_1.js')
        bundle_1.build()

        assert self.get('out_1.js') == 'more(),define("foo",function(){});\n;'


class TestInvenioAssetsCleanCSSFilter(TempEnvironmentHelper):
    """"Test class checking the behaviour of CSS filter """
    default_files = {
        'foo.css': """
        /* A comment */
            h1  {
                font-family: "Verdana"  ;
                 color: #FFFFFF;
            }
        """
    }

    def test_clean_CSS(self):
        """Test method of Clean CSS filter."""
        app = Flask(__name__)
        FlaskCLI(app)
        InvenioAssets(app)
        with app.app_context():
            bundle = self.mkbundle('foo.css',
                                   filters=CleanCSSFilter(),
                                   output='out.css')
            bundle.build()
            assert self.get('out.css') == 'h1{font-family:Verdana;color:#FFF}'



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Click command-line interface for assets and collect."""

from __future__ import absolute_import, print_function

import json
import logging
import os

import click
from flask import current_app
from flask_cli import with_appcontext
from pkg_resources import DistributionNotFound, get_distribution

from .proxy import current_assets


#
# Assets commands
#
def _webassets_cmd(cmd):
    """Helper to run a webassets command."""
    from webassets.script import CommandLineEnvironment
    logger = logging.getLogger('webassets')
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.DEBUG)
    cmdenv = CommandLineEnvironment(current_app.jinja_env.assets_environment,
                                    logger)
    getattr(cmdenv, cmd)()


@click.command()
@click.option("-i", "--bower-json", help="base input file", default=None,
              type=click.File('r'))
@click.option("-o", "--output-file", help="write bower.json to output file",
              metavar="FILENAME", type=click.File('w'))
@with_appcontext
def bower(bower_json, output_file):
    """Generate a bower.json file."""
    try:
        version = get_distribution(current_app.name).version
    except DistributionNotFound:
        version = ""

    output = {
        "name": current_app.name,
        "version": version,
        "dependencies": {},
    }

    # Load base file
    if bower_json:
        output = dict(output, **json.load(bower_json))

    # Iterate over bundles
    for bundle in current_app.extensions['invenio-assets'].env:
        if hasattr(bundle, 'bower'):
            output['dependencies'].update(bundle.bower)

    # Write to instance folder if output file is not specified
    if output_file is None:
        if not os.path.exists(current_app.instance_path):
            os.makedirs(current_app.instance_path)
        output_file = open(
            os.path.join(current_app.instance_path, "bower.json"),
            "w")

    click.echo("Writing {0}".format(output_file.name))
    json.dump(output, output_file, indent=4)
    output_file.close()

    # Write .bowerrc
    bowerrc = os.path.join(
        os.path.dirname(output_file.name),
        ".bowerrc")

    if not os.path.exists(bowerrc):
        click.echo("Writing {0}".format(bowerrc))

        static_path = os.path.relpath(
            os.path.join(current_app.static_folder, "bower_components"),
            current_app.instance_path
        )

        with open(bowerrc, 'w') as f:
            f.write(json.dumps(dict(directory=static_path)))


@click.group()
def assets():
    """Web assets commands."""


@assets.command()
@with_appcontext
def build():
    """Build bundles."""
    _webassets_cmd('build')


@assets.command()
@with_appcontext
def clean():
    """Clean bundles."""
    _webassets_cmd('clean')


@assets.command()
@with_appcontext
def watch():
    """Watch bundles for file changes."""
    _webassets_cmd('watch')


#
# Collect commands
#
@click.command()
@click.option('-v', '--verbose', default=False, is_flag=True)
@with_appcontext
def collect(verbose=False):
    """Collect static files."""
    current_assets.collect.collect(verbose=verbose)
